@@ -1,8 +1,11 @@
-//! This crate adds support for auto deriving ``bevy_asset_loader::AssetCollection``
+//! This crate adds support for auto deriving [bevy_asset_loader::AssetCollection]
 //!
 //! You do not have to use it directly. Just import ``AssetCollection`` from ``bevy_asset_loader``
 //! and use ``#[derive(AssetCollection)]`` to derive the trait.
 
+#![forbid(unsafe_code)]
+#![warn(unused_imports)]
+
 extern crate proc_macro;
 
 mod assets;
@@ -18,7 +21,8 @@ use syn::{Data, Field, Fields, Lit, Meta, NestedMeta};
 
 /// Derive macro for AssetCollection
 ///
-/// The helper attribute ``asset`` can be used to define the path to the asset file.
+/// The helper attribute ``asset`` can be used to define the path to the asset file
+/// and other asset options.
 #[proc_macro_derive(AssetCollection, attributes(asset))]
 pub fn asset_collection_derive(input: TokenStream) -> TokenStream {
     let ast = syn::parse(input).unwrap();
@@ -27,11 +31,12 @@ pub fn asset_collection_derive(input: TokenStream) -> TokenStream {
         .into()
 }
 
-const ASSET_ATTRIBUTE: &str = "asset";
-const PATH_ATTRIBUTE: &str = "path";
+pub(crate) const ASSET_ATTRIBUTE: &str = "asset";
+pub(crate) const PATH_ATTRIBUTE: &str = "path";
+pub(crate) const KEY_ATTRIBUTE: &str = "key";
 
-const TEXTURE_ATLAS_ATTRIBUTE: &str = "texture_atlas";
-struct TextureAtlasAttribute;
+pub(crate) const TEXTURE_ATLAS_ATTRIBUTE: &str = "texture_atlas";
+pub(crate) struct TextureAtlasAttribute;
 impl TextureAtlasAttribute {
     pub const TILE_SIZE_X: &'static str = "tile_size_x";
     pub const TILE_SIZE_Y: &'static str = "tile_size_y";
@@ -41,80 +46,8 @@ impl TextureAtlasAttribute {
     pub const PADDING_Y: &'static str = "padding_y";
 }
 
-const COLOR_MATERIAL_ATTRIBUTE: &str = "color_material";
-
-#[derive(Default)]
-struct AssetBuilder {
-    field_ident: Option<Ident>,
-    asset_path: Option<String>,
-    tile_size_x: Option<f32>,
-    tile_size_y: Option<f32>,
-    columns: Option<usize>,
-    rows: Option<usize>,
-    padding_x: f32,
-    padding_y: f32,
-    is_color_material: bool,
-}
-
-impl AssetBuilder {
-    fn build(self) -> Result<Asset, Vec<ParseFieldError>> {
-        let mut missing_fields = vec![];
-        if self.tile_size_x.is_none() {
-            missing_fields.push(format!(
-                "{}/{}",
-                TEXTURE_ATLAS_ATTRIBUTE,
-                TextureAtlasAttribute::TILE_SIZE_X
-            ));
-        }
-        if self.tile_size_y.is_none() {
-            missing_fields.push(format!(
-                "{}/{}",
-                TEXTURE_ATLAS_ATTRIBUTE,
-                TextureAtlasAttribute::TILE_SIZE_Y
-            ));
-        }
-        if self.columns.is_none() {
-            missing_fields.push(format!(
-                "{}/{}",
-                TEXTURE_ATLAS_ATTRIBUTE,
-                TextureAtlasAttribute::COLUMNS
-            ));
-        }
-        if self.rows.is_none() {
-            missing_fields.push(format!(
-                "{}/{}",
-                TEXTURE_ATLAS_ATTRIBUTE,
-                TextureAtlasAttribute::ROWS
-            ));
-        }
-        if self.field_ident.is_none() || self.asset_path.is_none() {
-            return Err(vec![ParseFieldError::NoAttributes]);
-        }
-        if missing_fields.len() == 4 {
-            let asset = BasicAsset {
-                field_ident: self.field_ident.unwrap(),
-                asset_path: self.asset_path.unwrap(),
-            };
-            if self.is_color_material {
-                return Ok(Asset::ColorMaterial(asset));
-            }
-            return Ok(Asset::Basic(asset));
-        }
-        if missing_fields.is_empty() {
-            return Ok(Asset::TextureAtlas(TextureAtlasAsset {
-                field_ident: self.field_ident.unwrap(),
-                asset_path: self.asset_path.unwrap(),
-                tile_size_x: self.tile_size_x.unwrap(),
-                tile_size_y: self.tile_size_y.unwrap(),
-                columns: self.columns.unwrap(),
-                rows: self.rows.unwrap(),
-                padding_x: self.padding_x,
-                padding_y: self.padding_y,
-            }));
-        }
-        Err(vec![ParseFieldError::MissingAttributes(missing_fields)])
-    }
-}
+pub(crate) const FOLDER_ATTRIBUTE: &str = "folder";
+pub(crate) const COLOR_MATERIAL_ATTRIBUTE: &str = "color_material";
 
 fn impl_asset_collection(
     ast: syn::DeriveInput,
@@ -136,6 +69,18 @@ fn impl_asset_collection(
                                 ParseFieldError::NoAttributes => {
                                     default_fields.push(field.clone().ident.unwrap())
                                 }
+                                ParseFieldError::EitherSingleAssetOrFolder => {
+                                    compile_errors.push(syn::Error::new_spanned(
+                                        field.into_token_stream(),
+                                        "You can only specify one of 'folder' or 'path'",
+                                    ));
+                                }
+                                ParseFieldError::KeyAttributeStandsAlone => {
+                                    compile_errors.push(syn::Error::new_spanned(
+                                        field.into_token_stream(),
+                                        "The 'key' attribute has to be the only 'asset' attribute on a field",
+                                    ));
+                                }
                                 ParseFieldError::MissingAttributes(missing_attributes) => {
                                     compile_errors.push(syn::Error::new_spanned(
                                         field.into_token_stream(),
@@ -188,12 +133,22 @@ fn impl_asset_collection(
         Asset::Basic(basic) => {
             let field_ident = basic.field_ident.clone();
             let asset_path = basic.asset_path.clone();
-            quote!(#es#field_ident : asset_server.get_handle(#asset_path),)
+            quote!(#es #field_ident : asset_server.get_handle(#asset_path),)
+        }
+        Asset::Folder(basic) => {
+            let field_ident = basic.field_ident.clone();
+            let asset_path = basic.asset_path.clone();
+            quote!(#es #field_ident : asset_server.load_folder(#asset_path).unwrap(),)
+        }
+        Asset::DynamicAsset(dynamic) => {
+            let field_ident = dynamic.field_ident.clone();
+            let asset_key = dynamic.key.clone();
+            quote!(#es #field_ident : asset_server.get_handle(asset_keys.get_path_for_key(#asset_key.into())),)
         }
         Asset::ColorMaterial(basic) => {
             let field_ident = basic.field_ident.clone();
             let asset_path = basic.asset_path.clone();
-            quote!(#es#field_ident : materials.add(asset_server.get_handle(#asset_path).into()),)
+            quote!(#es #field_ident : materials.add(asset_server.get_handle(#asset_path).into()),)
         }
         Asset::TextureAtlas(texture_asset) => {
             let field_ident = texture_asset.field_ident.clone();
@@ -205,7 +160,7 @@ fn impl_asset_collection(
             let padding_x = texture_asset.padding_x;
             let padding_y = texture_asset.padding_y;
             quote!(
-                #es#field_ident : {
+                #es #field_ident : {
                 atlases.add(TextureAtlas::from_grid_with_padding(
                     asset_server.get_handle(#asset_path),
                     Vec2::new(#tile_size_x, #tile_size_y),
@@ -218,7 +173,7 @@ fn impl_asset_collection(
     });
     asset_creation.append_all(default_fields.iter().fold(
         quote!(),
-        |es, ident| quote! (#es#ident : Default::default()),
+        |es, ident| quote! (#es #ident : Default::default()),
     ));
 
     let asset_loading = assets.iter().fold(quote!(), |es, asset| match asset {
@@ -226,6 +181,14 @@ fn impl_asset_collection(
             let asset_path = asset.asset_path.clone();
             quote!(#es handles.push(asset_server.load_untyped(#asset_path));)
         }
+        Asset::Folder(asset) => {
+            let asset_path = asset.asset_path.clone();
+            quote!(#es asset_server.load_folder(#asset_path).unwrap().drain(..).for_each(|handle| handles.push(handle));)
+        }
+        Asset::DynamicAsset(dynamic) => {
+            let asset_key = dynamic.key.clone();
+            quote!(#es handles.push(asset_server.load_untyped(asset_keys.get_path_for_key(#asset_key.into())));)
+        }
         Asset::ColorMaterial(asset) => {
             let asset_path = asset.asset_path.clone();
             quote!(#es handles.push(asset_server.load_untyped(#asset_path));)
@@ -239,23 +202,22 @@ fn impl_asset_collection(
     #[allow(unused_mut)]
     let mut conditional_asset_collections = quote! {};
 
-    #[cfg(feature = "sprite")]
+    #[cfg(feature = "render")]
     {
+        // color materials
         conditional_asset_collections = quote! {
         #conditional_asset_collections
                 let mut materials = cell
                     .get_resource_mut::<Assets<ColorMaterial>>()
-                    .expect("Cannot get Assets<ColorMaterial>");
+                    .expect("Cannot get resource Assets<ColorMaterial>");
         };
-    }
 
-    #[cfg(feature = "render")]
-    {
+        // texture atlas
         conditional_asset_collections = quote! {
         #conditional_asset_collections
                 let mut atlases = cell
                     .get_resource_mut::<Assets<TextureAtlas>>()
-                    .expect("Cannot get Assets<TextureAtlas>");
+                    .expect("Cannot get resource Assets<TextureAtlas>");
         };
     }
 
@@ -263,6 +225,7 @@ fn impl_asset_collection(
             fn create(world: &mut World) -> Self {
                 let cell = world.cell();
                 let asset_server = cell.get_resource::<AssetServer>().expect("Cannot get AssetServer");
+                let asset_keys = cell.get_resource::<bevy_asset_loader::AssetKeys>().expect("Cannot get bevy_asset_loader::AssetKeys");
                 #conditional_asset_collections
                 #name {
                     #asset_creation
@@ -270,18 +233,24 @@ fn impl_asset_collection(
             }
     };
 
+    let load_function = quote! {
+            fn load(world: &mut World) -> Vec<HandleUntyped> {
+                let cell = world.cell();
+                let asset_server = cell.get_resource::<AssetServer>().expect("Cannot get AssetServer");
+                let asset_keys = cell.get_resource::<bevy_asset_loader::AssetKeys>().expect("Cannot get bevy_asset_loader::AssetKeys");
+                let mut handles = vec![];
+                #asset_loading
+                handles
+            }
+    };
+
     let impl_asset_collection = quote! {
         #[automatically_derived]
         #[allow(unused_variables)]
         impl AssetCollection for #name {
             #create_function
 
-            fn load(world: &mut World) -> Vec<HandleUntyped> {
-                let asset_server = world.get_resource::<AssetServer>().expect("Cannot get AssetServer");
-                let mut handles = vec![];
-                #asset_loading
-                handles
-            }
+            #load_function
         }
     };
     Ok(impl_asset_collection)
@@ -289,6 +258,8 @@ fn impl_asset_collection(
 
 enum ParseFieldError {
     NoAttributes,
+    EitherSingleAssetOrFolder,
+    KeyAttributeStandsAlone,
     WrongAttributeType(proc_macro2::TokenStream, &'static str),
     UnknownAttributeType(proc_macro2::TokenStream),
     UnknownAttribute(proc_macro2::TokenStream),
@@ -307,11 +278,29 @@ fn parse_field(field: &Field) -> Result<Asset, Vec<ParseFieldError>> {
             for attribute in asset_meta_list.nested.iter() {
                 if let NestedMeta::Meta(Meta::NameValue(ref named_value)) = attribute {
                     let path = named_value.path.get_ident().unwrap().clone();
+                    builder.field_ident = Some(field.clone().ident.unwrap());
 
                     if path == PATH_ATTRIBUTE {
                         if let Lit::Str(path_literal) = &named_value.lit {
                             builder.asset_path = Some(path_literal.value());
-                            builder.field_ident = Some(field.clone().ident.unwrap());
+                        } else {
+                            errors.push(ParseFieldError::WrongAttributeType(
+                                named_value.clone().into_token_stream(),
+                                "str",
+                            ));
+                        }
+                    } else if path == FOLDER_ATTRIBUTE {
+                        if let Lit::Str(path_literal) = &named_value.lit {
+                            builder.folder_path = Some(path_literal.value());
+                        } else {
+                            errors.push(ParseFieldError::WrongAttributeType(
+                                named_value.clone().into_token_stream(),
+                                "str",
+                            ));
+                        }
+                    } else if path == KEY_ATTRIBUTE {
+                        if let Lit::Str(path_literal) = &named_value.lit {
+                            builder.key = Some(path_literal.value());
                         } else {
                             errors.push(ParseFieldError::WrongAttributeType(
                                 named_value.clone().into_token_stream(),
